# coding: utf-8
import inspect
import logging
import importlib
from django.conf import settings
from celery import group, chain
from celery import current_app
from .choices import INNER_STATE_RUNNING
from .models import AvailableTask
from .tasks import BaseTask, ChangeStateTask


logger = logging.getLogger(__name__)


def is_subclass(o):
    return inspect.isclass(o) and issubclass(o, BaseTask)


class AvailableTaskLoader(object):

    def _get_subclasses(self):
        task_dict = {}
        for app in settings.INSTALLED_APPS:
            try:
                mod = __import__('%s.%s' % (app, 'tasks'))
            except:
                continue
            members = inspect.getmembers(mod.tasks, predicate=lambda x: inspect.isclass(x) and issubclass(x, BaseTask))
            for m in members:
                # {'foo.bar.Task': 'class <foo.bar.Task>'}
                task_dict['{0}.{1}'.format(m[1].__module__, m[0])] = m[1]
        return task_dict

    def load(self):
        subcls = self._get_subclasses()
        for cls_name, cls in subcls.items():

            current_app.tasks.register(cls)

            if not hasattr(cls, 'public') or not cls.public:
                continue
            full_name = '{0}.{1}'.format(cls.__module__, cls.__name__)
            name = cls.name if cls.name else cls.__name__
            description = cls.description if cls.description else cls.__doc__
            try:
                obj, created = AvailableTask.objects.get_or_create(name=name,
                                                                   klass=full_name,
                                                                   description=description)
                logger.info('AvailableTask %s created successfully.', full_name)
            except Exception as ex:
                logger.warning('Error while creating %s. %s',
                               full_name,
                               ex.message)

        self._prune()

    def _prune(self):
        '''removes all the unecessary tasks'''
        tasks = set([member for member in self._get_subclasses().keys()])
        existing = set(AvailableTask.objects.all().values_list('klass', flat=True))
        stale = existing - tasks
        for s in stale:
            at = AvailableTask.objects.get(klass=s)
            at.transition_tasks.all().delete()
            at.delete()


class TaskLoader(object):

    def load_task(self, task):

        parts = task.split('.')

        try:
            module_name = '.'.join(parts[:-1])
            klass_name = parts[-1]
            module = importlib.import_module(module_name)
            klass = getattr(module, klass_name)
            return klass
        except:
            raise ValueError('Cannot load task {0}'.format(task))


class TaskRunner(object):

    def __init__(self, controller, next, task_loader=None):

        '''initializes a task'''

        if not controller:
            raise ValueError('Controller cannot be null.')

        if not next:
            raise ValueError('Next state cannot be null')

        self.task_loader = task_loader if task_loader is not None else TaskLoader()

        # supports passing controlled or controller objects
        if hasattr(controller, 'controller'):
            self.controller = controller.controller
            self.controlled = controller
        else:
            self.controller = controller
            self.controlled = controller.content_object

        self.next = next
        self.transition = self.get_transition()
        self.initialize_tasks()

    def get_transition(self):
        transitions = self.controller.machine.transitions.filter(from_state=self.controller.current_state,
                                                                 to_state=self.next)
        if transitions.count() != 1:
            raise ValueError('invalid transition for this taskrunner')

        return transitions.first()

    def initialize_tasks(self):

        '''loads and initializes all the tasks'''
        self.tasks = [self.task_loader.load_task(t.klass)()
                      for t in self.transition.tasks.all()]
        self.tasks.append(ChangeStateTask())

        self.validation_tasks = [self.task_loader.load_task(t.validation_class)()
                                 for t in self.tasks if t.validation_class and
                                 t.validation_class != '']

    def run(self):

        '''executes the tasks'''
        cid = self.controller.id
        nsi = self.next.id

        validation = group([v.s(cid=cid, nsi=nsi)
                            for v in self.validation_tasks])

        tasks = chain([t.s(cid=cid, nsi=nsi)
                       for t in self.tasks])
        if len(validation.tasks) > 0:
            job = chain(validation, tasks)
        else:
            job = tasks

        self.controller.inner_state = INNER_STATE_RUNNING
        self.controller.save()
        return job.delay()



# coding: utf-8
from copy import deepcopy
import celery
from django.core.exceptions import ObjectDoesNotExist
from rest_framework.decorators import detail_route, list_route
from rest_framework.response import Response
from rest_framework import viewsets, status
from rest_framework.permissions import IsAuthenticated
from common.viewsets import DefaultViewSetMixIn
from .choices import INNER_STATE_RUNNING
from .rest import INVALID_REQUEST
from .filters import (StateMachineFilter,
                      ActionFilter,
                      TransitionLogFilter,
                      AvailableTaskFilter,)
from .models import (StateMachine,
                     State,
                     Action,
                     Transition,
                     TransitionLog,
                     AvailableTask,
                     TransitionTask, 
                     StateController)
from .serializers import (StateMachineSerializer,
                          StateSerializer,
                          ActionSerializer,
                          AvailableTaskSerializer,
                          TransitionSerializer,
                          TransitionLogSerializer,
                          TransitionTaskSerializer, )


class CeleryInspectViewSet(DefaultViewSetMixIn,
                           viewsets.ViewSet):

    permission_classes = (IsAuthenticated, )
    inspect = celery.current_app.control.inspect()

    @list_route(methods=["get"])
    def ping(self, request):
        result = self.inspect.ping()
        return Response(result, status=status.HTTP_200_OK)

    @list_route(methods=["get"])
    def registered(self, request):
        result = self.inspect.registered()
        return Response(result, status=status.HTTP_200_OK)

    @list_route(methods=["get"])
    def active(self, request):
        result = self.inspect.active()
        new_result = []
        for k, v in result.items():
            new_result.extend(v)
        return Response(new_result, status=status.HTTP_200_OK)

    @list_route(methods=["get"])
    def scheduled(self, request):
        result = self.inspect.scheduled()
        return Response(result, status=status.HTTP_200_OK)


class ActionViewSet(DefaultViewSetMixIn,
                    viewsets.ModelViewSet):

    queryset = Action.objects.all()
    serializer_class = ActionSerializer
    filter_class = ActionFilter
    search_fields = ('id', 'name', )


class AvailableTaskViewSet(DefaultViewSetMixIn,
                           viewsets.ModelViewSet):

    queryset = AvailableTask.objects.all()
    serializer_class = AvailableTaskSerializer
    filter_class = AvailableTaskFilter
    search_fields = ('id', 'name', )


class StateMachineViewSet(DefaultViewSetMixIn,
                          viewsets.ModelViewSet):

    queryset = StateMachine.objects.all()
    serializer_class = StateMachineSerializer
    filter_class = StateMachineFilter
    search_fields = ('id', 'name', )

    def destroy(self, *args, **kwargs):
        try:
            StateController.objects.get(machine=self.get_object())
            return Response({'message': 'State machine can not be deleted if it has control'},
                            status=status.HTTP_409_CONFLICT)
        except ObjectDoesNotExist: 
            return super(DefaultViewSetMixIn, self).destroy(*args, **kwargs)


class StateViewSet(DefaultViewSetMixIn,
                   viewsets.ModelViewSet):

    queryset = State.objects.all()
    serializer_class = StateSerializer


class TransitionViewSet(DefaultViewSetMixIn,
                        viewsets.ModelViewSet):

    queryset = Transition.objects.all()
    serializer_class = TransitionSerializer


class TaskViewSet(DefaultViewSetMixIn,
                  viewsets.ModelViewSet):

    queryset = TransitionTask.objects.all()
    serializer_class = TransitionTaskSerializer


class TransitionLogViewSet(DefaultViewSetMixIn,
                           viewsets.ModelViewSet):

    queryset = TransitionLog.objects.all()
    serializer_class = TransitionLogSerializer
    filter_class = TransitionLogFilter
    search_fields = ('controller', )


class StateControllerViewSetMixIn(object):

    @detail_route(methods=['get', 'put'])
    def data(self, request, pk=None):
        self.controlled = self.get_object()
        if request.method == 'GET':
            data = self.controlled.current_data
            return Response(data.data, status=status.HTTP_200_OK)
        else:
            try:
                current_data = self.controlled.current_data
                current_data.data = request.data
                original = current_data.data
                copied = deepcopy(original)
                copied.update(request.data)
                current_data.data = copied
                current_data.save()
                return Response(request.data, status=status.HTTP_200_OK)
            except:
                return Response({'message': 'Fail on saving data.'},
                                status=status.HTTP_500_INTERNAL_SERVER_ERROR)

    @detail_route(methods=['post'])
    def change(self, request, pk=None):
        controlled = self.get_object()
        controller = controlled.controller
        if not controller:
            return INVALID_REQUEST

        if controller.inner_state == INNER_STATE_RUNNING:
            return Response({'status': 'FSM already running for this project.'},
                            status=status.HTTP_400_BAD_REQUEST)

        state_id = request.data.get('state_id', None)
        if not state_id:
            return INVALID_REQUEST
        try:
            state = State.objects.get(pk=state_id)
        except:
            return Response({'status': 'State does not exist'},
                            status=status.HTTP_400_BAD_REQUEST)
        available = controlled.next_for_user(request.user)
        if int(state_id) not in [s.to_state.id for s in available]:
            return Response({'status': 'Invalid transition'},
                            status=status.HTTP_400_BAD_REQUEST)

        task = controller.change_to(state)
        task_id = task.parent.id if task.parent else task.id

        return Response({
            'status': 'State change requested',
            'task': task_id
        })



# coding: utf-8
import json
from django.contrib.auth.models import Permission
from rest_framework.serializers import ValidationError
from .models import (StateController,
                     Action,
                     State,
                     Transition,
                     AvailableTask,
                     TransitionTask,)


class FSMUpdater(object):

    '''Updates the FSM underlying data
    based on a JSON representation'''

    def update(self, instance, data):
        raise NotImplementedError


class GoFSMUpdater(FSMUpdater):

    # issues #69
    # def validate_controlled(self, instance, data):

    #     if not instance.representation:
    #         return True

    #     old_representation = json.loads(instance.representation)
    #     new_representation = json.loads(data['representation'])

    #     old_nodes = set([node['text'] for node in old_representation['nodeDataArray']])
    #     new_nodes = set([node['text'] for node in new_representation['nodeDataArray']])

    #     stale = old_nodes - new_nodes
    #     if StateController.objects.filter(machine=instance,
    #                                       current_state__code__in=stale).count() > 0:
    #         raise ValidationError('You cannot change the FSM because there are controlled objects in a removed stated.')

    #     return True

    def update_status(self, fsm, old, new):
        '''this method updates the representation
        of a fsm when the state name is changed'''
        representation = json.loads(fsm.representation)
        for node in representation['nodeDataArray']:
            if node['text'] == old:
                node['text'] = new

        fsm.representation = json.dumps(representation)
        fsm.save()

    def update(self, instance, data):
        if 'representation' not in data:
            return None

        instance.transitions.all().delete()
        representation = json.loads(data['representation'])
        states = {s['key']: s['text'] for s in representation['nodeDataArray']}

        for rep_transition in representation['linkDataArray']:
            from_node = rep_transition['from']
            to_node = rep_transition['to']

            try:
                from_state = State.objects.get(code=states[from_node])
            except:
                from_state = State.objects.create(code=states[from_node])
            try:
                to_state = State.objects.get(code=states[to_node])
            except:
                to_state = State.objects.create(code=states[to_node])

            transition = Transition.objects.create(name=rep_transition['text'],
                                                   machine=instance,
                                                   from_state=from_state,
                                                   to_state=to_state)
            if 'tasks' in rep_transition:
                tasks = [AvailableTask.objects.get(pk=int(t)) for t in rep_transition['tasks']]
                for task in tasks:
                    TransitionTask.objects.create(transition=transition,
                                                  task=task)

            if 'permissions' in rep_transition:
                permissions = [Permission.objects.get(pk=int(p))
                               for p in rep_transition['permissions']]
                for permission in permissions:
                    transition.permissions.add(permission)

        for state_rep in representation['nodeDataArray']:
            state_type = state_rep.get('type', 'common')
            state = State.objects.get(code=state_rep['text'])
            state.actions.clear()
            for action_id in state_rep.get('actions', list()):
                try:
                    action = Action.objects.get(id=action_id)
                    state.actions.add(action)
                except Action.DoesNotExist:
                    pass
            if state_type == 'initial':
                instance.initial_state = state
                instance.save()

        if not instance.initial_state:
            raise ValidationError(u'You need to define an initial state for this FSM.')

        return instance



# coding: utf-8
import rest_framework_filters
from .models import (StateMachine,
                     Action,
                     AvailableTask,
                     TransitionLog,)


class ActionFilter(rest_framework_filters.FilterSet):

    class Meta:
        model = Action
        fields = {
            'id': ['exact', 'in'],
            'name': ['icontains']
        }


class AvailableTaskFilter(rest_framework_filters.FilterSet):

    class Meta:
        model = AvailableTask
        fields = {
            'id': ['exact', 'in'],
            'name': ['icontains']
        }


class StateMachineFilter(rest_framework_filters.FilterSet):

    class Meta:
        model = StateMachine
        fields = {
            'id': ['exact'],
            'name': ['icontains']
        }


class TransitionLogFilter(rest_framework_filters.FilterSet):

    class Meta:

        model = TransitionLog
        fields = {
            'controller': ['exact']
        }



# coding: utf-8
from django.utils.translation import ugettext_lazy as _
from rest_framework.response import Response
from rest_framework import status


INVALID_REQUEST = Response({'status': _('Invalid Request')},
                           status=status.HTTP_400_BAD_REQUEST)



# coding: utf-8
from django.utils.translation import ugettext_lazy as _
from django.contrib.auth.models import Permission
from django.contrib.gis.db import models
from django.contrib.postgres.fields import JSONField
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from ordered_model.models import OrderedModel
from common.models import (
    DomainModelMixIn,
    DateCreatedMixIn,
    DateUpdatedMixIn,
    CreatedByMixIn,
    CompiledDescriptionMixIn,
)
from .choices import (
    INNER_STATE_CHOICES,
    INNER_STATE_IDLE,
    INNER_STATE_RUNNING,
)
from .signals import (
    before_state_change,
    after_state_change,
    initialize_state_machine,
)


class StateMachine(DateCreatedMixIn,
                   DateUpdatedMixIn,
                   CreatedByMixIn,
                   CompiledDescriptionMixIn):

    name = models.CharField(verbose_name=_('Name'),
                            max_length=64)

    initial_state = models.ForeignKey('workflow.State',
                                      verbose_name=_('Initial State'),
                                      null=True)

    representation = JSONField(verbose_name=_('Representation'),
                               help_text=_('Graphic representation of the FSM graph in JSON.'),
                               null=True)

    def next(self, current_state):
        return self.transitions.filter(from_state=current_state)

    def next_for_user(self, current_state, user):
        transitions = self.next(current_state)
        return [t for t in transitions if t.is_available(user)]

    def __unicode__(self):

        return unicode(self.name)

    class Meta:

        verbose_name = _('State Machine')
        verbose_name_plural = _('State Machines')


class Action(DateCreatedMixIn,
             DateUpdatedMixIn,
             CreatedByMixIn):

    name = models.CharField(max_length=64,
                            verbose_name=_('Name'))

    def __unicode__(self):
        return self.name

    class Meta:

        verbose_name = _('Action')
        verbose_name_plural = _('Actions')


class State(DomainModelMixIn,
            CreatedByMixIn):

    actions = models.ManyToManyField(Action)

    class Meta:

        verbose_name = _('State')
        verbose_name_plural = _('States')


class Transition(DateCreatedMixIn,
                 DateUpdatedMixIn,
                 CreatedByMixIn):

    name = models.CharField(max_length=64,
                            verbose_name=_('Name'))

    machine = models.ForeignKey(StateMachine,
                                related_name='transitions')

    from_state = models.ForeignKey(State,
                                   related_name='from_transitions',
                                   on_delete=models.PROTECT)

    to_state = models.ForeignKey(State,
                                 related_name='to_transitions',
                                 on_delete=models.PROTECT)

    permissions = models.ManyToManyField(Permission)

    tasks = models.ManyToManyField('workflow.AvailableTask',
                                   through='workflow.TransitionTask',
                                   related_name='transitions')

    def is_available(self, user):
        '''determines if this transition can
        be executed by this user'''
        if self.permissions.all().count() <= 0:
            return True

        perms = ["{0}.{1}".format(p.content_type.app_label, p.codename)
                 for p in self.permissions.all()]

        return user.has_perms(perms)

    def __unicode__(self):

        return u'{0} - {1} > {2}'.format(self.machine,
                                         self.from_state,
                                         self.to_state)

    class Meta:

        unique_together = (('machine', 'from_state', 'to_state'), )
        verbose_name = _('Transition')
        verbose_name_plural = _('Transitions')


class AvailableTask(DateCreatedMixIn,
                    DateUpdatedMixIn,
                    CompiledDescriptionMixIn):

    name = models.CharField(verbose_name=_('Name'),
                            max_length=128)

    klass = models.CharField(verbose_name=_('Task'),
                             max_length=512)

    def __unicode__(self):
        return self.name

    class Meta:
        verbose_name = _('Available Task')
        verbose_name_plural = _('Available Tasks')


class TransitionTask(DateCreatedMixIn,
                     DateUpdatedMixIn,
                     CreatedByMixIn,
                     OrderedModel):

    transition = models.ForeignKey(Transition,
                                   related_name='transition_tasks',
                                   on_delete=models.CASCADE)

    task = models.ForeignKey(AvailableTask,
                             related_name='transition_tasks',
                             on_delete=models.CASCADE)

    def __unicode__(self):

        return unicode(self.task.klass)

    class Meta(OrderedModel.Meta):

        ordering = ('transition', 'order', )
        verbose_name = _('Task')
        verbose_name_plural = _('Tasks')


class StateController(models.Model):

    content_type = models.ForeignKey(ContentType)

    object_id = models.PositiveIntegerField()

    content_object = GenericForeignKey('content_type', 'object_id')

    machine = models.ForeignKey(StateMachine,
                                verbose_name=_('State Machine'))

    current_state = models.ForeignKey(State,
                                      verbose_name=_('Current State'),
                                      on_delete=models.PROTECT)

    task_id = models.CharField(max_length=64,
                               verbose_name=_('Task ID'),
                               help_text=_('Task ID that this controller is currently running.'),
                               null=True,
                               blank=True)

    inner_state = models.CharField(verbose_name=_('Inner State'),
                                   choices=INNER_STATE_CHOICES,
                                   default=INNER_STATE_IDLE,
                                   max_length=32)

    def next(self):

        return self.machine.transitions.filter(from_state=self.current_state)

    def can_change_to(self, next):
        '''Validates if it's a valid
        transition'''
        if self.inner_state != INNER_STATE_IDLE:
            return False

        # transitions = self.next().filter(to_state=next)

        # if transitions.count() != 1:
        #     # more then one transition is impossible
        #     return False

        return True

    def change_to(self, next):
        '''changes the state to the next one'''
        from .task_runner import TaskRunner
        if not self.can_change_to(next):
            return False

        before_state_change.send_robust(sender=self.__class__,
                                        controlled=self.content_object,
                                        controller=self,
                                        current=self.current_state,
                                        next=next)

        task_runner = TaskRunner(self, next)
        return task_runner.run()

    @property
    def current_data(self):
        return self.data.filter(state=self.current_state).latest('date_created')

    class Meta:

        unique_together = (('content_type', 'object_id'), )
        verbose_name = _('State Controller')
        verbose_name_plural = _('State Controllers')


class StateControllerData(DateCreatedMixIn,
                          DateUpdatedMixIn):

    controller = models.ForeignKey(StateController,
                                   verbose_name=_('State Controller'),
                                   related_name='data')

    state = models.ForeignKey(State,
                              verbose_name=_('State'),
                              on_delete=models.PROTECT)

    data = JSONField(verbose_name=_('State Data'),
                     default=dict)

    class Meta:

        ordering = ('-date_created', )


class StateControllerMixIn(object):

    def save(self, state_machine=None, *args, **kwargs):
        if not self.pk:
            new = True
        else:
            new = False
        super(StateControllerMixIn, self).save(*args, **kwargs)

        if new and state_machine:
            initialize_state_machine.send_robust(sender=self.__class__,
                                                 controlled=self,
                                                 state_machine=state_machine,
                                                 initial_state=state_machine.initial_state)

    @property
    def controller(self):
        content_type = ContentType.objects.get_for_model(self.__class__)
        try:
            sc = StateController.objects.get(content_type_id=content_type.id,
                                             object_id=self.id)
        except:
            return None

        return sc

    @property
    def current_data(self):
        if self.controller:
            return self.controller.current_data

    @property
    def current_state(self):
        if self.controller:
            return self.controller.current_state

        return None

    @property
    def next(self):
        if self.controller:
            return self.controller.machine.next(self.current_state)

        return None

    def next_for_user(self, user):
        if self.controller and self.controller.machine:
            return self.controller.machine.next_for_user(self.current_state,
                                                         user)
        return None

    def can_change_to(self, next):
        '''Validates if it's a valid
        transition'''
        return self.controller.can_change_to(next)

    def change_to(self, next):
        '''Changes the state machine
        to a new state and fires all the stuff it
        needs to do'''
        return self.controller.change_to(next)


class TransitionLog(DateCreatedMixIn):

    controller = models.ForeignKey(StateController,
                                   verbose_name=_('State Controller'),
                                   related_name='transition_logs')

    from_state = models.ForeignKey(State,
                                   verbose_name=_('From State'),
                                   related_name='+',
                                   null=True)

    to_state = models.ForeignKey(State,
                                 verbose_name=_('To State'),
                                 related_name='+',
                                 null=True)

    class Meta:

        verbose_name = _('Transition Log')
        verbose_name_plural = _('Transition Logs')
        ordering = ('-date_created', )



# coding: utf-8
import django.dispatch

initialize_state_machine = django.dispatch.Signal(providing_args=['controlled', 'state_machine', 'initial_state'])
before_state_change = django.dispatch.Signal(providing_args=['controlled', 'controller', 'current', 'next'])
after_state_change = django.dispatch.Signal(providing_args=['controlled', 'controller', 'previous', 'current'])



# coding: utf-8
import logging
from django.apps import AppConfig
from django.conf import settings
from reversion import revisions as reversion


logger = logging.getLogger(__name__)


class WorkflowApp(AppConfig):

    name = 'workflow'
    verbose_name = 'workflow'

    def ready(self):
        from workflow.receivers import *  # noqa
        from .models import (StateMachine,
                             State,
                             Transition,
                             TransitionTask,
                             AvailableTask,)
        reversion.register(StateMachine)
        reversion.register(State)
        reversion.register(TransitionTask)
        reversion.register(AvailableTask)

        if hasattr(settings, 'WORKFLOW_AUTO_LOAD') and settings.WORKFLOW_AUTO_LOAD:
            logging.debug('Autoloading tasks.')
            try:
                from .task_runner import AvailableTaskLoader
                atl = AvailableTaskLoader()
                atl.load()
            except:
                logging.debug('Autoloading failed. If this is a migration, dont worry')



# -*- coding: utf-8 -*-
from django.conf.urls import url
from django.views.generic import TemplateView

from . import views

urlpatterns = [
    url(r'', TemplateView.as_view(template_name="base.html")),
    ]



# coding: utf-8
from __future__ import absolute_import
import logging
from .models import (State,
                     StateController, )
from celery import current_app
from .choices import INNER_STATE_IDLE
from .signals import after_state_change
logger = logging.getLogger(__name__)


class BaseTask(current_app.Task):

    ignore_result = False
    validation_class = ''
    name = 'Base Task'
    description = ''

    def on_failure(self, exc, task_id, args, kwargs, einfo):
        logger.error(exc.message)
        if self.controller:
            self.controller.inner_state = INNER_STATE_IDLE
            self.controller.task_id = None
            self.controller.save()

    def run(self, *args, **kwargs):
        controller_id = kwargs.pop('cid', None)
        next_id = kwargs.pop('nsi', None)
        self._load_data(controller_id, next_id)
        task_id = self.request.get('root_id', None)
        if not task_id:
            task_id = self.request.get('id', None)
        self.controller.task_id = task_id
        self.controller.save()
        return self._run()

    def _load_data(self, controller_id, next_id):
        self.controller = StateController.objects.get(id=controller_id)
        self.previous = self.controller.current_state
        self.next = State.objects.get(id=next_id)

    def _run(self):

        return True


class ValidateSchemaTask(BaseTask):

    schema = None

    def _run(self):

        if self.schema:
            errors = self.schema.validate(self.controller.current_data.data)
            if len(errors) > 0:
                raise ValueError('Schema Invalid')

        return True


class ChangeStateTask(BaseTask):

    name = 'Change State'
    description = 'Changes the current state to the next.'
    public = False

    def _run(self):
        self.controller.inner_state = INNER_STATE_IDLE
        self.controller.current_state = self.next
        self.controller.task_id = None
        self.controller.save()
        after_state_change.send_robust(sender=self.controller.content_object.__class__,
                                       controlled=self.controller.content_object,
                                       controller=self.controller,
                                       previous=self.previous,
                                       current=self.next)
        return True



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-05-31 18:26
from __future__ import unicode_literals

from django.conf import settings
import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('auth', '0008_alter_user_username_max_length'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='Action',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
                ('date_updated', models.DateTimeField(auto_now=True, verbose_name='Date Updated')),
                ('name', models.CharField(max_length=64, verbose_name='Name')),
                ('created_by', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='User')),
            ],
            options={
                'verbose_name': 'Action',
                'verbose_name_plural': 'Actions',
            },
        ),
        migrations.CreateModel(
            name='AvailableTask',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.TextField(null=True, verbose_name='Description')),
                ('_cached_description', models.TextField(null=True, verbose_name='Cached Description')),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
                ('date_updated', models.DateTimeField(auto_now=True, verbose_name='Date Updated')),
                ('name', models.CharField(max_length=128, verbose_name='Name')),
                ('klass', models.CharField(max_length=512, verbose_name='Task')),
            ],
            options={
                'verbose_name': 'Available Task',
                'verbose_name_plural': 'Available Tasks',
            },
        ),
        migrations.CreateModel(
            name='State',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
                ('date_updated', models.DateTimeField(auto_now=True, verbose_name='Date Updated')),
                ('code', models.CharField(max_length=32, unique=True, verbose_name='Code')),
                ('description', models.CharField(max_length=128, null=True, verbose_name='Description')),
                ('actions', models.ManyToManyField(to='workflow.Action')),
                ('created_by', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='User')),
            ],
            options={
                'verbose_name': 'State',
                'verbose_name_plural': 'States',
            },
        ),
        migrations.CreateModel(
            name='StateController',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('object_id', models.PositiveIntegerField()),
                ('task_id', models.CharField(blank=True, help_text='Task ID that this controller is currently running.', max_length=64, null=True, verbose_name='Task ID')),
                ('inner_state', models.CharField(choices=[(b'idle', 'Idle'), (b'running', 'Running')], default=b'idle', max_length=32, verbose_name='Inner State')),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
                ('current_state', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='workflow.State', verbose_name='Current State')),
            ],
            options={
                'verbose_name': 'State Controller',
                'verbose_name_plural': 'State Controllers',
            },
        ),
        migrations.CreateModel(
            name='StateControllerData',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
                ('date_updated', models.DateTimeField(auto_now=True, verbose_name='Date Updated')),
                ('data', django.contrib.postgres.fields.jsonb.JSONField(default=dict, verbose_name='State Data')),
                ('controller', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='data', to='workflow.StateController', verbose_name='State Controller')),
                ('state', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='workflow.State', verbose_name='State')),
            ],
            options={
                'ordering': ('-date_created',),
            },
        ),
        migrations.CreateModel(
            name='StateMachine',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.TextField(null=True, verbose_name='Description')),
                ('_cached_description', models.TextField(null=True, verbose_name='Cached Description')),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
                ('date_updated', models.DateTimeField(auto_now=True, verbose_name='Date Updated')),
                ('name', models.CharField(max_length=64, verbose_name='Name')),
                ('representation', django.contrib.postgres.fields.jsonb.JSONField(help_text='Graphic representation of the FSM graph in JSON.', null=True, verbose_name='Representation')),
                ('created_by', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='User')),
                ('initial_state', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='workflow.State', verbose_name='Initial State')),
            ],
            options={
                'verbose_name': 'State Machine',
                'verbose_name_plural': 'State Machines',
            },
        ),
        migrations.CreateModel(
            name='Transition',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
                ('date_updated', models.DateTimeField(auto_now=True, verbose_name='Date Updated')),
                ('name', models.CharField(max_length=64, verbose_name='Name')),
                ('created_by', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='User')),
                ('from_state', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='from_transitions', to='workflow.State')),
                ('machine', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='transitions', to='workflow.StateMachine')),
                ('permissions', models.ManyToManyField(to='auth.Permission')),
            ],
            options={
                'verbose_name': 'Transition',
                'verbose_name_plural': 'Transitions',
            },
        ),
        migrations.CreateModel(
            name='TransitionLog',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
                ('controller', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='transition_logs', to='workflow.StateController', verbose_name='State Controller')),
                ('from_state', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='workflow.State', verbose_name='From State')),
                ('to_state', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='workflow.State', verbose_name='To State')),
            ],
            options={
                'ordering': ('-date_created',),
                'verbose_name': 'Transition Log',
                'verbose_name_plural': 'Transition Logs',
            },
        ),
        migrations.CreateModel(
            name='TransitionTask',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
                ('date_updated', models.DateTimeField(auto_now=True, verbose_name='Date Updated')),
                ('order', models.PositiveIntegerField(db_index=True, editable=False)),
                ('created_by', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='User')),
                ('task', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='transition_tasks', to='workflow.AvailableTask')),
                ('transition', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='transition_tasks', to='workflow.Transition')),
            ],
            options={
                'ordering': ('transition', 'order'),
                'abstract': False,
                'verbose_name': 'Task',
                'verbose_name_plural': 'Tasks',
            },
        ),
        migrations.AddField(
            model_name='transition',
            name='tasks',
            field=models.ManyToManyField(related_name='transitions', through='workflow.TransitionTask', to='workflow.AvailableTask'),
        ),
        migrations.AddField(
            model_name='transition',
            name='to_state',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='to_transitions', to='workflow.State'),
        ),
        migrations.AddField(
            model_name='statecontroller',
            name='machine',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='workflow.StateMachine', verbose_name='State Machine'),
        ),
        migrations.AlterUniqueTogether(
            name='transition',
            unique_together=set([('machine', 'from_state', 'to_state')]),
        ),
        migrations.AlterUniqueTogether(
            name='statecontroller',
            unique_together=set([('content_type', 'object_id')]),
        ),
    ]
